
from .version import __version__, __title__, __description__, __url__, \
    __author__, __author_email__, __license__, __copyright__


#from . import lib
#from . import collections, ec, memory, stream
#from .pattern_32a import allocate_pattern_buffer, PatternTx, PatternRx
