# Copyright 2021 Jetperch LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import dateutil.parser
import datetime


FBP_TIME_EPOCH_UNIX_OFFSET_SECONDS = dateutil.parser.parse('2018-01-01T00:00:00Z').timestamp()
Q_GAIN = 2 ** 30


def now():
    k = datetime.datetime.utcnow().timestamp()
    rel_s = k - FBP_TIME_EPOCH_UNIX_OFFSET_SECONDS
    return int(rel_s * Q_GAIN)
